#include <assert.h>
#include <ctype.h>
#include <netdb.h>
#include <netinet/in.h>
#include <pwd.h>
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/un.h>
#include <stdlib.h>
#include "common.h"
#include "list.h"

#define TA_BYTE(x)	((x) >> 3)
#define TA_MASK(x)	(0x80 >> ((x) & 0x07))
#define CHANNEL_INC	1
#define GAME_INC	1
#define CLIENT_INC	16

void                     init();
int                      daemonInit();
void                     processConfig();
void                     SIGTERM_handler(int sig);
int                      serverListen();
void                     channelAlloc();
void                     gameAlloc();
void                     clientAlloc();
int                      clientAdd(int fd);
void                     clientDel(int fd, fd_set *allset, fd_set *rset);
int                      channelNo(char *name);
int                      clientNo(char *name);
void                     writeRes(int c, int resType, int reslen, void *res);
void                     writeRes2(int c, int resType, int len1, void *res1, int len2, void *res2);
void                     request(int c, char *cmdP);
void                     connectClientToNetwork(int c);
void                     disconnectClientFromNetwork(int c);
void                     expireAutoMsgs();
void                     writeAutoMsgs(int c, int chan);
time_t                   addTime(char *s);
void                     writeChannels(int c);
void                     writeClients(int c, int chan);
void                     logQuit(char *fmt,...);

time_t                   startTime;
List                     autoMsg;
int                      nClients;
Client                  *client;
int                      nChannels;
Channel                 *channel;
int                      nGames;
Game                    *game;

int
main()
{
  int                      i,
                           j,
                           maxFD,
                           maxI,
                           serverFD,
                           cliFD;
  word                     n, nRead;
  char                     buf[MAX_CMDLEN + 16];
  struct sockaddr_in       addr;
  word                     len = sizeof(addr);
  time(&startTime);
  fd_set                   rSet,
                           allSet;
  time_t                   nextRun = time(NULL) + 60;

  init();
  FD_ZERO(&allSet);
  if ((serverFD = serverListen()) < 0)
    logQuit("serverListen error");
  FD_SET(serverFD, &allSet);
  maxFD = serverFD;
  maxI = -1;
  for (;;) {
    rSet = allSet;
    if (select(maxFD + 1, &rSet, NULL, NULL, NULL) < 0)
      logQuit("select error");
    if (time(NULL) >= nextRun) {
      /* recalculate maxFD, maxI for efficiency */
      maxFD = serverFD;
      maxI = -1;
      for (i = 0; i < nClients; i++)
	if (client[i].fd >= 0) {
	  maxFD = max(maxFD, client[i].fd);
	  maxI = i;
	}

      /* expire auto-messages */
      expireAutoMsgs();

      /* ping clients */
      n = MSG_PING;
      for (i = 0; i < nClients; i++)
	if ((cliFD = client[i].fd) >= 0) {
	  if (client[i].lastPingSent - client[i].lastPingRcvd > 60)
	    clientDel(cliFD, &allSet, &rSet);
	  else {
	    writeRes(i, n, sizeof(int), &n);
	    client[i].lastPingSent = time(NULL);
	  }
	}
      nextRun = time(NULL) + 60;
    }
    if (FD_ISSET(serverFD, &rSet)) {
      if ((cliFD = accept(serverFD, (struct sockaddr *) &addr, &len)) < 0)
	logQuit("accept() error");
      i = clientAdd(cliFD);
      FD_SET(cliFD, &allSet);
      maxFD = max(maxFD, cliFD);
      maxI = max(maxI, i);
    }
    for (i = 0; i <= maxI; i++) {
      if ((cliFD = client[i].fd) < 0)
	continue;
      if (FD_ISSET(cliFD, &rSet)) {
	nRead = 0;
	while (nRead < sizeof(int)) {
	  j = read(cliFD, (char*)&n + nRead, sizeof(int) - nRead);
	  if (j <= 0) {
	    clientDel(cliFD, &allSet, &rSet);
	    goto l0;
	  }
	  nRead += j;
	}
	if (n > MAX_CMDLEN) {
	  clientDel(cliFD, &allSet, &rSet);
	  goto l0;
	}
	nRead = 0;
	while (nRead < n) {
	  n = read(cliFD, buf + nRead, n - nRead);
	  if (n <= 0) {
	    clientDel(cliFD, &allSet, &rSet);
	    goto l0;
	  }
	  nRead += n;
	}
	for (j = MAX_CMDLEN; j < MAX_CMDLEN + 16; j++)
	  buf[j] = '\0';
	request(i, buf);
l0:
	continue;
      }
    }
  }
}

void
init()
{
  openlog("lgcd", 0, LOG_DAEMON);
  signal(SIGPIPE, SIG_IGN);
  signal(SIGTERM, SIGTERM_handler);
  channel = (Channel *) malloc(MAX_CHANNELS * sizeof(Channel));
  assert(channel != NULL);
  processConfig();
  listInit(&autoMsg, strcmp, 0, TRUE);
  daemonInit();
}

int
daemonInit()
{
  if (fork()) exit(0);
  for (int i = 0; i <= 2; i++)
    close(i);
  setsid();
  chdir("/");
  umask(0077);
  return 0;
}

void
processConfig()
{
  int                      i,
                           li;
  FILE                    *f;
  char                     line[256],
                           token[256];

  f = fopen(LIB "/lgcd.config", "rt");
  assert(f != NULL);
  while (fgets(line, 256, f) != NULL) {
    li = 0;
    if (!getToken(token, line, &li)) {
      if (!strcasecmp(token, "#"))
	continue;
      else if (!strcasecmp(token, "Channel")) {
	channelAlloc();
	if (getToken(token, line, &li))
	  continue;
	strcpy(channel[nChannels].name, token);
	i = 0;
	while (line[li++] != '\"');
	while (line[li] != '\"')
	  channel[nChannels].topic[i++] = line[li++];
	channel[nChannels].topic[i] = '\0';
	channel[nChannels].nConnects = 0;
	channel[nChannels].nClients = 0;
	listInit(&channel[nChannels].autoMsg, strcmp, 0, TRUE);
	li++;
	if (getToken(token, line, &li))
	  continue;
	channel[nChannels].maxNet = atoi(token);
	if (getToken(token, line, &li))
	  continue;
	channel[nChannels].gameNo = atoi(token);
	if (!getToken(token, line, &li))
	  channel[nChannels].dynamic = !strcasecmp(token, "DYNAMIC");
	nChannels++;
      } else if (!strcasecmp(token, "Game")) {
	gameAlloc();
	if (getToken(token, line, &li))
	  continue;
	game[nGames].clientNo = atoi(token);
	while (isspace(line[++li]));
	strcpy(game[nGames].setup, line + li);
	i = strlen(game[nGames].setup);
	if (i)
	  game[nGames].setup[i - 1] = '\0';
	nGames++;
      }
    }
  }
  fclose(f);
}

void
SIGTERM_handler(int sig)
{
  sig = sig;
  logQuit("caught a SIGTERM.");
  exit(0);
}

int
serverListen()
{
  int                      fd;
  struct sockaddr_in       addr;
  char                     hostname[256];
  struct hostent          *hP;

  if ((fd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
    perrQuit("socket");
  bzero(&addr, sizeof(addr));
  addr.sin_family = AF_INET;
  gethostname(hostname, 256);
  hP = gethostbyname(hostname);
  memcpy(&addr.sin_addr, hP->h_addr, hP->h_length);
  addr.sin_port = htons(2001);
  if (bind(fd, (struct sockaddr *) &addr, sizeof(addr)))
    perrQuit("bind");
  if (listen(fd, 5) < 0)
    return -1;
  return fd;
}

void
channelAlloc()
{
  if (!channel)
    channel = (Channel *) malloc(CHANNEL_INC * sizeof(Channel));
  else
    channel = (Channel *) realloc(channel, (nChannels + CLIENT_INC) * sizeof(Channel));
}

void
gameAlloc()
{
  if (!game)
    game = (Game *) malloc(GAME_INC * sizeof(Game));
  else
    game = (Game *) realloc(game, (nGames + CLIENT_INC) * sizeof(Game));
}

void
clientAlloc()
{
  int                      i;

  if (!client)
    client = (Client *) malloc(CLIENT_INC * sizeof(Client));
  else
    client = (Client *) realloc(client, (nClients + CLIENT_INC) * sizeof(Client));
  for (i = nClients; i < nClients + CLIENT_INC; i++)
    client[i].fd = -1;
  nClients += CLIENT_INC;
}

int
clientAdd(int fd)
{
  int                      i;

  if (!client)
    clientAlloc();
again:
  for (i = 0; i < nClients; i++) {
    if (client[i].fd == -1) {
      client[i].fd = fd;
      client[i].channel = -1;
      client[i].busy = 0;
      client[i].ccid = -1;
      client[i].nodeAddr = 0;
      client[i].lastPingSent = 0;
      client[i].lastPingRcvd = 0;
      client[i].name[0] = '\0';
      client[i].gecos[0] = '\0';
      client[i].line[0] = '\0';
      return i;
    }
  }
  clientAlloc();
  goto again;
}

void
clientDel(int fd, fd_set *allset, fd_set *rset)
{
  int                      i,
                           j,
                           res;
  List                    *list;
  AutoMsg                 *autoMsg;

  for (i = 0; i < nClients; i++) {
    if (client[i].fd == fd) {
      for (j = -1; j < nChannels; j++) {
	list = (j == -1) ? &::autoMsg : &channel[j].autoMsg;
	listIterateInit(list, HEAD_TO_TAIL);
	while ((autoMsg = (AutoMsg *) listIterate(list, NULL)) != NULL)
	  autoMsg->seen[TA_BYTE(i)] &= ~TA_MASK(i);
      }
      if (client[i].name[0])
	for (j = 0; j < nClients; j++)
	  if (client[j].fd >= 0 && j != i) {
	    res = MSG_SIGNOFF;
	    writeRes(j, res, strlen(client[i].name) + 1, client[i].name);
	  }
      if (client[i].channel >= 0) {
	channel[client[i].channel].nClients--;
	if (client[i].ccid >= 0)
	  disconnectClientFromNetwork(i);
	for (j = 0; j < nClients; j++) {
	  if (client[j].fd >= 0 && j != i) {
	    if (client[j].ccid == i) {
	      client[j].busy = 0;
	      client[j].ccid = -1;
	      channel[client[i].channel].nConnects--;
	    }
	  }
	}
      }
      FD_CLR(client[i].fd, allset);
      FD_CLR(client[i].fd, rset);
      close(client[i].fd);
      client[i].fd = -1;
      break;
    }
  }
}

int
channelNo(char *name)
{
  int                      i,
                           n = -1,
                           len = strlen(name);

  for (i = 0; i < nChannels; i++)
    if (!strncasecmp(channel[i].name, name, len)) {
      if (strlen(channel[i].name) == (word) len)
	return i;
      if (n >= 0)
	return -1;
      else
	n = i;
    }
  return n;
}

int
clientNo(char *name)
{
  int                      i,
                           n = -1,
                           len = strlen(name);

  for (i = 0; i < nClients; i++)
    if (client[i].fd >= 0 && !strncasecmp(client[i].name, name, len)) {
      if (strlen(client[i].name) == (word) len)
	return i;
      if (n >= 0)
	return -1;
      else
	n = i;
    }
  return n;
}

void
writeRes(int c, int resType, int reslen, void *res)
{
  write(client[c].fd, &resType, sizeof(int));
  write(client[c].fd, &reslen, sizeof(int));
  write(client[c].fd, res, reslen);
}

void
writeRes2(int c, int resType, int len1, void *res1, int len2, void *res2)
{
  int                      len = len1 + len2;
  write(client[c].fd, &resType, sizeof(int));
  write(client[c].fd, &len, sizeof(int));
  write(client[c].fd, res1, len1);
  write(client[c].fd, res2, len2);
}

void
request(int c, char *cmdP)
{
  int                      i,
                           j,
                           k,
                           cmd,
                           addr[MAX_NETNODES],
                           len1,
                           len2,
                           res = RES_OK;
  byte                     buf[256];
  ServerInfo               serverInfo;
  AutoMsg                 *autoMsg;
  List                    *list;
  ListNode                *node;
  time_t                   expire;
  char                    *s;

  switch ((cmd = *(int *) cmdP)) {
  case CMD_PING:
    client[c].lastPingRcvd = time(NULL);
    break;
/* -------------------------------------------------------------------------------------------- */
  case CMD_SERVERINFO:
    serverInfo.startTime = startTime;
    serverInfo.upTime = time(NULL) - startTime;
    serverInfo.maxNet = MAX_CLIENTS;
    serverInfo.nChannels = nChannels;
    for (j = 0, i = 0; i < nClients; i++)
      if (client[i].fd >= 0)
	j++;
    serverInfo.nClients = j;
    for (j = 0, i = 0; i < nClients; i++)
      if (client[i].fd >= 0 && client[i].ccid >= 0)
	j++;
    serverInfo.netConnections = j;
    for (j = 0, i = 0; i < nClients; i++)
      if (client[i].fd >= 0 && client[i].ccid >= 0 && client[client[i].ccid].ccid == client[i].ccid)
	j++;
    serverInfo.activeNetConnections = j;
    for (j = 0, i = 0; i < nClients; i++)
      if (client[i].fd >= 0 && client[i].ccid == i)
	j++;
    serverInfo.activeGames = j;
    writeRes(c, res, sizeof(ServerInfo), &serverInfo);
    break;
/* -------------------------------------------------------------------------------------------- */
  case CMD_CLIENTINFO:
    s = cmdP + sizeof(int);
    if (clientNo(s) >= 0) {
      res = RES_EBADNICK;
      writeRes(c, res, sizeof(int), &c);
      break;
    }
    strncpy(client[c].name, s, 12); client[c].name[12] = '\0';
    s += strlen(s) + 1;
    if (strlen(s) > 31) break;
    strncpy(client[c].gecos, s, 31); client[c].gecos[31] = '\0';
    s += strlen(s) + 1;
    s[UT_LINESIZE - 1] = '\0';
    strncpy(client[c].line, s, UT_LINESIZE - 1); client[c].line[UT_LINESIZE - 1] = '\0';
    writeRes(c, res, sizeof(int), &c);
    for (i = 0; i < nClients; i++)
      if (client[i].fd >= 0 && i != c) {
	res = MSG_SIGNON;
	writeRes(i, res, strlen(client[c].name) + 1, client[c].name);
      }
    break;
/* -------------------------------------------------------------------------------------------- */
  case CMD_NICKNAME:
    len1 = strlen(client[c].name) + 1;
    len2 = strlen(cmdP + sizeof(int)) + 1;
    for (i = sizeof(int); cmdP[i]; i++)
      if (!isalpha(cmdP[i]) && !isdigit(cmdP[i])) {
	res = RES_EBADNICK;
	writeRes(c, res, sizeof(int), &res);
	return;
      }
    for (i = 0; i < nClients; i++) {
      if (client[i].fd >= 0 && i != c && !strcasecmp(client[i].name, (char *) (cmdP + sizeof(int)))) {
	res = RES_ENICKINUSE;
	writeRes(c, res, sizeof(int), &res);
	return;
      }
    }
    writeRes(c, res, sizeof(int), &res);
    for (i = 0; i < nClients; i++)
      if (client[i].fd >= 0 && !client[i].busy)
	writeRes2(i, MSG_NICKNAME, len1, client[c].name, len2, cmdP + sizeof(int));
    strncpy(client[c].name, cmdP + sizeof(int), 12); client[c].name[12] = '\0';
    break;
/* -------------------------------------------------------------------------------------------- */
  case CMD_GETCHANNELNO:
    i = channelNo(cmdP + sizeof(int));
    writeRes(c, res, sizeof(int), &i);
    break;
/* -------------------------------------------------------------------------------------------- */
  case CMD_LISTCHANNELS:
    writeChannels(c);
    break;
/* -------------------------------------------------------------------------------------------- */
  case CMD_LISTCLIENTS:
    writeClients(c, -1);
    break;
/* -------------------------------------------------------------------------------------------- */
  case CMD_JOINCHANNEL:
    i = channelNo(cmdP + sizeof(int));
    if (i < 0) {
      res = RES_ENOCHANNEL;
      writeRes(c, res, sizeof(int), &res);
      break;
    }
    if (client[c].channel >= 0) {
      for (j = 0; j < nClients; j++) {
	if (client[j].fd >= 0 && j != c && client[j].channel == client[c].channel &&
	    !client[j].busy)
	  writeRes(j, MSG_CLIENTLEFT, strlen(client[c].name) + 1, client[c].name);
      }
      if (client[c].ccid >= 0) {
	client[c].ccid = -1;
	channel[client[c].channel].nConnects--;
      }
      channel[client[c].channel].nClients--;
    }
    client[c].channel = i;
    channel[i].nClients++;
    res = RES_OK;
    writeRes(c, res, sizeof(int), &i);
    for (i = 0; i < nClients; i++) {
      if (client[i].fd >= 0 && i != c && client[i].channel == client[c].channel && !client[i].busy)
	writeRes(i, MSG_CLIENTJOINED, strlen(client[c].name) + 1, client[c].name);
    }
    break;
/* -------------------------------------------------------------------------------------------- */
  case CMD_LEAVECHANNEL:
    if (client[c].channel < 0 || client[c].busy) {
      res = RES_EILLEGAL;
      writeRes(c, res, sizeof(int), &res);
      break;
    }
    for (i = 0; i < nClients; i++) {
      if (client[i].fd >= 0 && i != c && client[i].channel == client[c].channel &&
	  !client[i].busy)
	writeRes(i, MSG_CLIENTLEFT, strlen(client[c].name) + 1, client[c].name);
    }
    if (client[c].ccid >= 0) {
      client[c].ccid = -1;
      channel[client[c].channel].nConnects--;
    }
    channel[client[c].channel].nClients--;
    client[c].channel = -1;
    writeRes(c, res, sizeof(int), &res);
    break;
/* -------------------------------------------------------------------------------------------- */
  case CMD_ACTION:
    if (client[c].channel == -1) {
      res = RES_EILLEGAL;
      writeRes(c, res, sizeof(int), &res);
      break;
    }
    writeRes(c, res, sizeof(int), &res);
    len1 = strlen(client[c].name) + 1;
    len2 = strlen(cmdP + sizeof(int)) + 1;
    for (i = 0; i < nClients; i++)
      if (client[i].fd >= 0 && client[i].channel == client[c].channel && !client[i].busy)
	writeRes2(i, MSG_ACTION, len1, client[c].name, len2, cmdP + sizeof(int));

    break;
/* -------------------------------------------------------------------------------------------- */
  case CMD_AUTOMSG:
    s = cmdP + sizeof(int);
    if (!strcmp(s, "*")) {
      i = -1;
    } else {
      i = channelNo(s);
      if (i < 0) {
	res = RES_EILLEGAL;
	writeRes(c, res, sizeof(int), &res);
	break;
      }
    }
    s += strlen(s) + 1;
    expire = addTime(s);
    if (!expire || expire > time(NULL) + 24 * 60 * 60) {
      res = RES_ENOCHANNEL;
      writeRes(c, res, sizeof(int), &res);
      break;
    }
    autoMsg = (AutoMsg *) malloc(sizeof(AutoMsg));
    strcpy(autoMsg->fromName, client[c].name);
    s += strlen(s) + 1;
    strncpy(autoMsg->toName, s, 12); autoMsg->toName[12] = '\0';
    s += strlen(s) + 1;
    strcpy(autoMsg->msg, s);
    bzero(autoMsg->seen, MAX_CLIENTS / 8);
    autoMsg->expire = expire;
    if (i == -1)
      listAddTail(&::autoMsg, autoMsg);
    else
      listAddTail(&channel[i].autoMsg, autoMsg);
    writeRes(c, res, sizeof(int), &res);
    break;
/* -------------------------------------------------------------------------------------------- */
  case CMD_AUTODEL:
    s = cmdP + sizeof(int);
    len1 = strlen(s);
    for (i = -1; i < nChannels; i++) {
    list = (i == -1) ? &::autoMsg : &channel[i].autoMsg;
      listIterateInit(list, HEAD_TO_TAIL);
      while ((autoMsg = (AutoMsg *) listIterate(list, &node)) != NULL) {
	if (!strcasecmp(autoMsg->fromName, client[c].name) && (!strcmp(s, "*") || !strncasecmp(autoMsg->msg, s, len1)))
	  listDel(list, node);
      }
    }
    writeRes(c, res, sizeof(int), &res);
    break;
/* -------------------------------------------------------------------------------------------- */
  case CMD_GETAUTOMSGS:
    if (!strcmp(cmdP + sizeof(int), "."))
      i = client[c].channel;
    else {
      i = channelNo(cmdP + sizeof(int));
      if (i < 0) {
	if (client[c].busy) {
	  res = RES_EILLEGAL;
	  writeRes(c, res, sizeof(int), &res);
	  break;
	}
      }
    }
    writeAutoMsgs(c, i);
    break;
/* -------------------------------------------------------------------------------------------- */
  case CMD_MSGUSER:
    i = clientNo(cmdP + sizeof(int));
    if (i == c) {
      res = RES_EILLEGAL;
      writeRes(c, res, sizeof(int), &res);
      break;
    }
    if (i < 0) {
      res = RES_ENOUSER;
      writeRes(c, res, sizeof(int), &res);
      break;
    }
    if (client[i].busy) {
      res = RES_EBUSY;
      writeRes(c, res, sizeof(int), &res);
      break;
    }
    writeRes(c, res, sizeof(int), &res);
    len1 = strlen(client[i].name) + 1;
    s = cmdP + sizeof(int);
    s += strlen(s) + 1;
    len2 = strlen(s) + 1;
    writeRes2(c, MSG_MSGSENT, len1, client[i].name, len2, s);
    len1 = strlen(client[c].name) + 1;
    writeRes2(i, MSG_PRIVMSG, len1, client[c].name, len2, s);
    break;
/* -------------------------------------------------------------------------------------------- */
  case CMD_MSGALL:
    if (client[c].channel >= 0) {
      len1 = strlen(client[c].name) + 1;
      len2 = strlen(cmdP + sizeof(int)) + 1;
      for (i = 0; i < nClients; i++)
	if (client[i].fd >= 0 && i != c && client[i].channel == client[c].channel && !client[i].busy)
	  writeRes2(i, MSG_BROADCAST, len1, client[c].name, len2, cmdP + sizeof(int));
    }
    writeRes(c, res, sizeof(int), &res);
    break;
/* -------------------------------------------------------------------------------------------- */
  case CMD_CONNECT:
    i = clientNo(cmdP + sizeof(int));
    if (i < 0) {
      res = RES_ENOUSER;
      writeRes(c, res, sizeof(int), &res);
      break;
    }
    if (client[c].channel < 0 || !channel[client[c].channel].maxNet) {
      res = RES_EILLEGAL;
      writeRes(c, res, sizeof(int), &res);
      break;
    }
    if (client[i].channel != client[c].channel) {
      res = RES_ENOTLOCAL;
      writeRes(c, res, sizeof(int), &res);
      break;
    }
    if (client[i].busy && !channel[client[c].channel].dynamic) {
      res = RES_ENETBUSY;
      writeRes(c, res, sizeof(int), &res);
      break;
    }
    for (k = 0, j = 0; j < nClients; j++)
      if (client[j].fd >= 0 && client[j].ccid == i)
	k++;
    if (i != c && client[i].ccid < 0)
      k++;
    if (k == channel[client[c].channel].maxNet) {
      res = RES_ENETFULL;
      writeRes(c, res, sizeof(int), &res);
      break;
    }
    for (k = 0, j = 0; j < nClients; j++)
      if (client[j].fd >= 0 && client[j].ccid >= 0)
	k++;
    if (i != c && client[i].ccid < 0)
      k++;
    if (k == MAX_CLIENTS) {
      res = RES_ENETFULL;
      writeRes(c, res, sizeof(int), &res);
      break;
    }
    if (i == c && !channel[client[c].channel].dynamic) {
      for (k = 1, j = 0; j < nClients; j++)
	if (client[j].fd >= 0 && j != c && client[j].ccid == c)
	  k++;
      if (k < 2) {
	res = RES_EILLEGAL;
	writeRes(c, res, sizeof(int), &res);
	break;
      }
    }
    res = RES_OK;
    writeRes(c, res, sizeof(int), &i);
    // send out informative messages
    for (j = 0; j < nClients; j++)
      if (client[j].fd >= 0 && client[j].channel == client[i].channel && !client[j].busy) {
	if (j == i) {
	  if (j != c)
	    writeRes(j, MSG_CONNREQ, strlen(client[c].name) + 1, client[c].name);
	} else
	  writeRes2(j, MSG_CLIENTCONN, strlen(client[c].name) + 1, client[c].name,
		    strlen(client[i].name) + 1, client[i].name);
      }
    // do the connection
    if (client[c].ccid < 0)
      channel[client[c].channel].nConnects++;
    client[c].ccid = i;
    if (i == c)
      client[c].nodeAddr = 0;
    else {
      bzero(addr, MAX_NETNODES * sizeof(int));
      for (j = 0; j < nClients; j++)
	if (client[j].fd >= 0 && j != c && client[j].ccid == i)
	  addr[client[j].nodeAddr] = 1;
      for (j = 1; j < MAX_NETNODES; j++)
	if (!addr[j]) {
	  client[c].nodeAddr = j;
	  break;
	}
    }
    if (i == c || (client[i].ccid == i && channel[client[c].channel].dynamic)) {
      res = MSG_CONNECT;
      for (k = 0, j = 0; j < nClients; j++)
	if (client[j].fd >= 0 && client[j].ccid == i)
	  k++;
      if (i == c) {
	buf[1] = 0;
	for (j = 0; j < nClients; j++)
	  if (client[j].fd >= 0 && client[j].ccid == c) {
	    buf[0] = game[channel[client[j].channel].gameNo].clientNo;
	    buf[1] = 0;
	    buf[2] = k;
	    buf[3] = client[j].nodeAddr;
	    strcpy((char *)buf + 4, game[channel[client[c].channel].gameNo].setup);
	    writeRes(j, res, 4 + strlen((char *) buf + 4) + 1, buf);
	  }
	for (j = 0; j < nClients; j++)
	  if (client[j].fd >= 0 && client[j].ccid == c)
	    connectClientToNetwork(j);
      } else {
	buf[0] = game[channel[client[c].channel].gameNo].clientNo;
	buf[1] = 0;
	buf[2] = k;
	buf[3] = client[c].nodeAddr;
	strcpy((char *)buf + 4, game[channel[client[c].channel].gameNo].setup);
	writeRes(c, res, 4 + strlen((char *) buf + 4), buf);
	connectClientToNetwork(c);
      }
    }
    break;
/* -------------------------------------------------------------------------------------------- */
  case CMD_DISCONNECT:
    i = client[c].ccid;
    if (i < 0) {
      res = RES_EILLEGAL;
      writeRes(c, res, sizeof(int), &res);
      break;
    }
    writeRes(c, res, sizeof(int), &res);
    for (j = 0; j < nClients; j++)
      if (client[j].fd >= 0 && client[j].channel == client[c].channel && !client[j].busy)
	writeRes2(j, MSG_CLIENTDISC, strlen(client[c].name) + 1, client[c].name,
		  strlen(client[i].name) + 1, client[i].name);
    disconnectClientFromNetwork(c);
    break;
  }
}

void
connectClientToNetwork(int c)
{
  int                      i;

  client[c].busy = 1;
  for (i = 0; i < nClients; i++)
    if (client[i].fd >= 0 && i != c && client[i].ccid == client[c].ccid) {
      writeRes2(c, MSG_CONNECT, sizeof(int), &client[i].nodeAddr, sizeof(int), &i);

      writeClients(c, -1);
      writeRes2(i, MSG_CONNECT, sizeof(int), &client[c].nodeAddr, sizeof(int), &c);

      writeClients(i, -1);
    }
}

void
disconnectClientFromNetwork(int c)
{
  int                      i,
                           j;

  if (client[c].busy)
    for (i = 0; i < nClients; i++)
      if (client[i].fd >= 0 && i != c && client[i].ccid == client[c].ccid)
	writeRes(i, MSG_DISCONNECT, sizeof(int), &client[c].nodeAddr);

  if (client[c].ccid == c) {			 // figure out who "owns" the net now...
    for (i = 0; i < nClients; i++)
      if (client[i].fd >= 0 && i != c && client[i].ccid == c) {
	for (j = 0; j < nClients; j++)
	  if (client[j].fd >= 0 && j != c && client[j].ccid == c)
	    client[j].ccid = i;
	break;
      }
  }
  client[c].busy = 0;
  client[c].ccid = -1;
  client[c].nodeAddr = -1;
  channel[client[c].channel].nConnects--;
}

void
expireAutoMsgs()
{
  int                      i;
  AutoMsg                 *autoMsg;
  List                    *list;
  ListNode                *node;

  for (i = -1; i < nChannels; i++) {
  list = (i == -1) ? &::autoMsg : &channel[i].autoMsg;
    listIterateInit(list, HEAD_TO_TAIL);
    while ((autoMsg = (AutoMsg *) listIterate(list, &node)) != NULL) {
      if (time(NULL) >= autoMsg->expire)
	listDel(list, node);
    }
  }
}

void
writeAutoMsgs(int c, int chan)
{
  int                      broadcast,
                           seen,
                           res = RES_OK;
  AutoMsg                 *autoMsg;
  List                    *list;
  ListNode                *node;
  writeRes(c, res, sizeof(int), &res);

  list = (chan == -1) ? &::autoMsg : &channel[chan].autoMsg;
  listIterateInit(list, HEAD_TO_TAIL);
  while ((autoMsg = (AutoMsg *) listIterate(list, &node)) != NULL) {
    seen = autoMsg->seen[TA_BYTE(c)] & TA_MASK(c);
    broadcast = !strcmp(autoMsg->toName, "*");
    if (!seen && (broadcast || !strcasecmp(autoMsg->toName, client[c].name))) {
      writeRes2(c, MSG_AUTOMSG, strlen(autoMsg->fromName) + 1, autoMsg->fromName, strlen(autoMsg->msg) + 1, autoMsg->msg);
      if (broadcast)
	autoMsg->seen[TA_BYTE(c)] |= TA_MASK(c);
      else
	listDel(list, node);
    }
  }
  res = MSG_DONEAUTOMSGS;
  writeRes(c, res, sizeof(int), &res);
}

time_t
addTime(char *s)
{
  time_t                   t = time(NULL),
                           endOfMinute,
                           endOfHour;
  struct tm               *tmP = localtime(&t);
  int                      len = strlen(s);
  char                     lastC = tolower(s[len - 1]);
  int                      n;

  n = atoi(s);
  if (n <= 0)
    return 0;
  endOfMinute = t + 60 - tmP->tm_sec;
  if (isdigit(lastC) || lastC == 'm')
    return endOfMinute + (n - 1) * 60;
  endOfHour = endOfMinute +
    (60 - tmP->tm_sec) +
    60 * (59 - tmP->tm_min);
  if (lastC == 'h')
    return endOfHour + (n - 1) * 60 * 60;
  return 0;
}

void
writeChannels(int c)
{
  int                      i,
                           fd = client[c].fd,
                           res = RES_OK;
  AutoMsg                 *autoMsg;
  write(fd, &res, sizeof(int));
  i = sizeof(int) + nChannels * sizeof(Channel);
  write(fd, &i, sizeof(int));
  write(fd, &nChannels, sizeof(int));
  for (i = 0; i < nChannels; i++) {
    channel[i].nAutoMsgs = 0;
    listIterateInit(&channel[i].autoMsg, HEAD_TO_TAIL);
    while ((autoMsg = (AutoMsg *) listIterate(&channel[i].autoMsg, NULL)) != NULL) {
      if (!(autoMsg->seen[TA_BYTE(c)] & TA_MASK(c)) &&
	  (!strcasecmp(autoMsg->toName, client[c].name) ||
	   !strcmp(autoMsg->toName, "*")))
	channel[i].nAutoMsgs++;
    }
    write(fd, &channel[i], sizeof(Channel));
  }
}

void
writeClients(int c, int chan)
{
  int                      i,
                           n,
                           fd = client[c].fd,
                           res = RES_OK;
  write(fd, &res, sizeof(int));
  for (n = 0, i = 0; i < nClients; i++)
    if (chan < 0 || (client[i].fd >= 0 && client[i].channel == chan))
      n++;
  i = sizeof(int) + n * sizeof(Client);
  write(fd, &i, sizeof(int));
  write(fd, &n, sizeof(int));
  for (i = 0; i < nClients; i++)
    if (chan < 0 || (client[i].fd >= 0 && client[i].channel == chan))
      write(fd, &client[i], sizeof(Client));
}

void
logQuit(char *fmt,...)
{
  char                     buf[256];
  va_list                  ap;

  va_start(ap, fmt);
  vsprintf(buf, fmt, ap);
  syslog(LOG_INFO, buf);
  exit(1);
}
